/*
 * @(#)feedback.c   3.0    3/27/86
 *
 * Draw and erase interactive feedback
 *
 * A feedback set is a figure consisting of straight lines connecting
 * a set of points.  If the figure is to be a polygon it must be 
 * explicitely closed (i.e. the last point must be the first).  The figure is
 * drawn directly onto the screen using pw_vector so that it happens fast.
 *
 */

#include "gremlin.h"
#include "icondata.h"

/* local definitions */
#define	ALL	1
#define LAST	2

/* imports from main.c */

extern Window		pix_sw;
extern struct rect	pix_size;

extern SUN_XORIGIN;
extern SUN_YORIGIN;

/* local static variables */

static int NPointsDisp;			/* number of points being displayed */
static struct pm_pos FBSet[MAXPOINTS];		/* vertices of the echo set */
static int Closed;			/* 1 if feedback is closed polygon */

/*
 * Circle drawing stuff
 */
	/* constants for drawing circles /*
	/* (actually just sin and cos of 0, pi/16, pi/8, 3pi/16, pi/4 ) */
static double sinx[5] = {0.0, .1950903, .3826834, .5555702, .7071068},
	      cosx[5] = {1.0, .9807853, .9238795, .8314696, .7071068};


/*
 *
 * Define a new set of points to feedback 
 */

FBDefine(point, opt)
POINT point;
int opt;
{
	register int i;
	register POINT *pt;

	if (NPointsDisp) FBErase(ALL);	/* erase any current feedback */
	for (pt = &point, i=0; !Nullpoint(pt); pt = PTNextPoint(pt), ++i) {
	    FBSet[i].x = dbx_to_win(pt->x); /* copy to feedback*/
	    FBSet[i].y = dby_to_win(pt->y);
	}

	NPointsDisp = i;
	Closed = opt;
	FBDisplay();			/* display the set */
}

/*
 * Erase all or part of the feedback set.  If option=LAST then just erase the
 * last line segment and point; otherwise erase the whole thing.
 */

FBErase(option)
int option;				/* option to FBErase */
{
	if (option == ALL) {
	    FBDisplay();		/* redisplay to erase */
	    NPointsDisp = 0;		/* nothing there */
	}
	else if (NPointsDisp > 1) {
	    FBDLine(NPointsDisp-2);	/* erase lines to the next to last */
	    FBDLine(NPointsDisp-1);	/* point */
	    if (Closed) FBDLine(NPointsDisp); /* erase the closing line */
	    NPointsDisp--;
	    FBSet[NPointsDisp-1].x = FBSet[NPointsDisp].x; /* move cursor */
	    FBSet[NPointsDisp-1].y = FBSet[NPointsDisp].y;
	    FBDLine(NPointsDisp-1);     /* redraw line to the cursor */
	    if (Closed) FBDLine(NPointsDisp);	/* reclose the figure */
	}
}

/*
 * Add a point/line to the feedback set
 */

int
FBAdd(point)
POINT point;
{
	if (NPointsDisp >= MAXPOINTS) return -1;
	if (Closed) FBDLine(NPointsDisp);	/* erase closing line */
	FBSet[NPointsDisp].x = dbx_to_win(point.x);
	FBSet[NPointsDisp].y = dby_to_win(point.y);
	NPointsDisp++;
	FBDLine(NPointsDisp-1);		/* draw the line */
	if (Closed) FBDLine(NPointsDisp);	/* close the polygon */
	return NPointsDisp;
}

/*
 * Change the location of a point in the feedback set
 */


FBChange(point,num)
POINT point;
int num;
{
	FBDLine(num);			/* erase the num-1 to num line */
	if ((num < NPointsDisp)	|| (Closed))	/* if not the last point */
		FBDLine(num+1);		/* erase the num to num+1 line */

	FBSet[num].x = dbx_to_win(point.x);
	FBSet[num].y = dby_to_win(point.y);

 	FBDLine(num);			/* draw the num-1 to num line */
	if ((num < NPointsDisp)	|| (Closed))	/* if not the last point */
		FBDLine(num+1);		/* draw the num to num+1 line */
}

/*
 * Echo a Box defined by upper and lower bounds.
 */

FBBox(xmax,ymax,xmin,ymin)
int xmax,ymax,xmin,ymin;
{
	if (NPointsDisp) FBErase(ALL);

	FBSet[0].x = FBSet[1].x = (short) xmin;
	FBSet[0].y = FBSet[3].y = (short) ymin;
	FBSet[2].x = FBSet[3].x = (short) xmax;
	FBSet[1].y = FBSet[2].y = (short) ymax;
	NPointsDisp = 4;
	Closed = 1;

	FBDisplay();
}

/*
 * Display the feedback set.
 */

FBDisplay()
{
	register int temp;

	if (NPointsDisp>1) {		/* if something to display */
#ifdef notdef
	    pw_lock(pix_pw, &pix_size);
#endif
	    for (temp=1; temp<NPointsDisp; temp++) {
		XDrawLine(display, pix_sw, pix_xor_gc,
		          FBSet[temp-1].x, FBSet[temp-1].y, 
			  FBSet[temp].x, FBSet[temp].y);
	    }
	    if ((Closed) && (NPointsDisp > 2)) {
		XDrawLine(display, pix_sw, pix_xor_gc,
			  FBSet[NPointsDisp-1].x, FBSet[NPointsDisp-1].y,
			  FBSet[0].x, FBSet[0].y);
	    }
#ifdef notdef
	    pw_unlock(pix_pw);
#endif
	}
}

/*
 * Draw a line from FBSet[num-1] to FBSet[num].
 */

FBDLine(num)
int num;
{
	if ((num >= NPointsDisp) || (num == 0)) /* maybe to close the figure */
	  if ((Closed)	&& (NPointsDisp > 2))	/* yes it is */
		/* note: never close a 2-sided figure - you get a blank line */
	    XDrawLine(display, pix_sw, pix_xor_gc, 
		      FBSet[NPointsDisp-1].x, FBSet[NPointsDisp-1].y,
	              FBSet[0].x, FBSet[0].y);
	  else return;				/* no - error */
	else				/* a valid interior request */
	  XDrawLine(display, pix_sw, pix_xor_gc,
	            FBSet[num-1].x, FBSet[num-1].y, 
		    FBSet[num].x, FBSet[num].y);
}


FBCircle(CenterX,CenterY,RadX,RadY)
int CenterX,CenterY,RadX,RadY;
{
	register int i, d1, d2;

	double rad;

	if (NPointsDisp) FBErase(ALL);

	d1 = CenterX-RadX;
	d2 = CenterY-RadY;
	rad = sqrt((double) d1*d1 + d2*d2);
	if (rad < 3.0) return;	/* don't bother with small ones */

	d2 = (int) (rad+0.5);

	FBSet[4].x = CenterX; FBSet[4].y = CenterY-d2;
	FBSet[12].x = CenterX-d2; FBSet[12].y = CenterY;
	FBSet[20].x = CenterX; FBSet[20].y = CenterY+d2;
	FBSet[28].x = CenterX+d2; FBSet[28].y = CenterY;

	for (i=1; i<4; i++) {
	    d1 = (int) (rad * sinx[i] + 0.5);
	    d2 = (int) (rad * cosx[i] + 0.5);
	    FBSet[4+i].x = CenterX-d1; FBSet[4+i].y = CenterY-d2;
	    FBSet[4-i].x = CenterX+d1; FBSet[4-i].y = CenterY-d2;
	    FBSet[12+i].x = CenterX-d2; FBSet[12+i].y = CenterY+d1;
	    FBSet[12-i].x = CenterX-d2; FBSet[12-i].y = CenterY-d1;
	    FBSet[20+i].x = CenterX+d1; FBSet[20+i].y = CenterY+d2;
	    FBSet[20-i].x = CenterX-d1; FBSet[20-i].y = CenterY+d2;
	    FBSet[28+i].x = CenterX+d2; FBSet[28+i].y = CenterY-d1;
	    FBSet[28-i].x = CenterX+d2; FBSet[28-i].y = CenterY+d1;
	}

	d1 = (int) (rad * sinx[4] + 0.5);
	d2 = (int) (rad * cosx[4] + 0.5);
	FBSet[0].x = CenterX+d1; FBSet[0].y = CenterY-d2;
	FBSet[8].x = CenterX-d2; FBSet[8].y = CenterY-d1;
	FBSet[16].x = CenterX-d1; FBSet[16].y = CenterY+d2;
	FBSet[24].x = CenterX+d2; FBSet[24].y = CenterY+d1;

	NPointsDisp = 32;
	Closed = 1;

	FBDisplay();
}
