/*
 * @(#)graphics.c	1.2	1/3/85
 *
 * Graphics routines for the SUN Gremlin picture editor.
 *
 * Original by Mark Opperman 
 *
 * Modifications by Yih-Farn Chen (yfchen@monet.berkeley.edu)
 * Modifications by Jim Thompson (james@ernie.berkeley.edu)
 *
 */

#include <math.h>
#include <sys/types.h>
#include "icondata.h"
#include "gremlin.h"
#include "stipple_patterns.h"

/* imports from main.c */

extern			error();
extern Window		pix_sw;
extern struct rect	pix_size;
extern PMRec		cset_pm;
extern PMRec		scratch_pm;
extern ELT *cset;
extern Artmode;
extern CSIZE;
extern CFONT;
extern CsetOn;
extern SUN_XORIGIN; 
extern SUN_YORIGIN; 

/* imports from display.c */

extern struct rect cset_rect;
extern struct pm_pos cset_orig;
extern minsunx, maxsunx, minsuny, maxsuny;

/* imports from C */

extern char *malloc();

/* forward references */

/* symbolic font from text.c */

XFontStruct *text_pf;

/* locally defined variables */

int charysizes[NFONTS][NSIZES];	/* Character y dimensions for each size */
int curve_set;			/* TRUE if spline points pre-computed */

int linestyle;			/* Current line style */
int linemod;			/* Type of line (SOLID, DOTTED, ...) */
int linethickness;		/* 1, 2, 3 */

char stippletype[32] = "cf";

/*
 * X font specifications.  Names of fonts include a "%d" which is later filled
 * in with the desired point size.
 */
char *font_types[NFONTS] = { "r", "i", "b", "s" };
char *font_names[NFONTS] = {
    "*-*-times-medium-r-*--%d-*-75-75-*-*-iso8859-*",
    "*-*-times-medium-i-*--%d-*-75-75-*-*-iso8859-*",
    "*-*-times-bold-r-*--%d-*-75-75-*-*-iso8859-*",
    "*-*-symbol-medium-r-*--%d-*-75-75-*-*-*-*"
    };
int font_sizes[NSIZES] = { 10, 12, 18, 24 };

int stipple_index[NMENUSTIPPLES] = { 1, 3, 12, 14, 16, 19, 21, 23 };
extern PMRec fill_pms[];

/* NOTE: all stipple fonts are expected to be 32 x 32 bit rasters */

/* pointers to the stipple pixmaps (16 x 16 bits) in the menu window */
PMRec *stipple_pms[NMENUSTIPPLES] = {
	&stipple1_pm, &stipple2_pm, &stipple3_pm, &stipple4_pm,
	&stipple5_pm, &stipple6_pm, &stipple7_pm, &stipple8_pm
};

/* the following pixmaps are defined in ../icons/icon.openbox" */
extern PMRec menu_stipple1_pm;
extern PMRec menu_stipple2_pm;
extern PMRec menu_stipple3_pm;
extern PMRec menu_stipple4_pm;
extern PMRec menu_stipple5_pm;
extern PMRec menu_stipple6_pm;
extern PMRec menu_stipple7_pm;
extern PMRec menu_stipple8_pm;

/* pointers to the stipple pixmaps (16 x 16 bits) in the pop-up menu of ETC */
PMRec *menu_stipple_pms[NMENUSTIPPLES] = {
	&menu_stipple1_pm, &menu_stipple2_pm, &menu_stipple3_pm,
	&menu_stipple4_pm, &menu_stipple5_pm, &menu_stipple6_pm,
	&menu_stipple7_pm, &menu_stipple8_pm
};

/* ... and the corresponding images (32 x 32 bits) */
char stipple_patterns[NDISPSTIPPLES][128];

/* data used in graphics2.c for drawing polygons */
int rasterlength;		/* real # horizontal bits in scratch_pm */
int bytesperline;		/* rasterlength / 8 */
int nlines;			/* # horizontal bits defined by scratch_pm */
char *fill;			/* pointer to scratch_pm image */

/* 
 *  This matrix points to the 16 fonts.
 */
XFontStruct *font_info[NFONTS][NSIZES] = {
    { (XFontStruct *)NULL, (XFontStruct *)NULL, (XFontStruct *)NULL, 
      (XFontStruct *)NULL },
    { (XFontStruct *)NULL, (XFontStruct *)NULL, (XFontStruct *)NULL, 
      (XFontStruct *)NULL },
    { (XFontStruct *)NULL, (XFontStruct *)NULL, (XFontStruct *)NULL, 
      (XFontStruct *)NULL },
    { (XFontStruct *)NULL, (XFontStruct *)NULL, (XFontStruct *)NULL, 
      (XFontStruct *)NULL },
};
PMRec char_pm;

/* Splines use these global arrays */

static float h[MAXPOINTS];
static float x[MAXPOINTS], dx[MAXPOINTS], d2x[MAXPOINTS], d3x[MAXPOINTS];
static float y[MAXPOINTS], dy[MAXPOINTS], d2y[MAXPOINTS], d3y[MAXPOINTS];
static numpoints;

/* These are used as bit masks to create the right style lines. */
#define SOLID -1
#define DOTTED 002
#define DASHED 004
#define DOTDASHED 012


/*
 * This routine sets the current line style.
 */
GRSetLineStyle(style)
int style;			/* new stipple pattern for lines */
{
    switch (linestyle = style) {
	case 1:			/* dotted */
	    linemod = DOTTED;
	    linethickness = 1;
	    break;
	case 2:			/* broken */
	    linemod = DOTDASHED;
	    linethickness = 1;
	    break;
	case 3:			/* thick */
	    linemod = SOLID;
	    linethickness = 3;
	    break;
	case 4:			/* dashed */
	    linemod = DASHED;
	    linethickness = 1;
	    break;
	case 5:			/* narrow */
	    linemod = SOLID;
	    linethickness = 1;
	    break;
	case 6:			/* medium */
	    linemod = SOLID;
	    linethickness = 2;
	    break;
    }
}


/*
 * This routine returns the maximum vertical size (in bits) of a character
 * of the specified font/size.
 */
GRGetCharYSize(font, size)
register font;			/* character font (1 - 4) */
register size;			/* character size (1 - 4) */
{
    return(charysizes[--font][--size]);
}


#define pi 3.14159265359
#define twopi 6.28318530718
#define log2_10 3.321915


/* 
 * Draw arc - always to scratch_pm.
 * Note: must check for zero radius before calling.
 */
GRArc(center, cpoint, angle, style)
register POINT *center, *cpoint;
float angle;
register style;
{
    double radius, resolution, t1, fullcircle;
    double degreesperpoint;
    float xs, ys, epsilon;
    float x1, y1, x2, y2;
    register i, extent;

    xs = cpoint->x - center->x;
    ys = cpoint->y - center->y;

    /* calculate drawing parameters */

    radius = sqrt((double) (xs * xs + ys * ys));
    t1 = floor(log10(radius) * log2_10);
    resolution = pow(2.0, t1);
    epsilon = (float) 1.0 / resolution;
    fullcircle = ceil(twopi * resolution);
    degreesperpoint = 360.0 / fullcircle;

    extent = (angle == 0) ? fullcircle : angle/degreesperpoint;

    GRSetLineStyle(style);

    if (angle == 0) {			/* do full circle fast */
	extent = max((int) (resolution/16), 4); /* number of lines per octet */
	degreesperpoint = twopi/(extent*8.0);	/* radians per line */
	xs = center->x;			/* center point */
	ys = center->y;
	x1 = radius*0.0;		/* sin of 0 */
	y1 = radius*1.0;		/* cos of 0 */
	for (i=1; i<=extent; ++i) {
	    x2 = radius*sin(i*degreesperpoint);
	    y2 = radius*cos(i*degreesperpoint);
	    GRVector(xs+x1, ys+y1, xs+x2, ys+y2);	/* draw symmetric */
	    GRVector(xs+x1, ys-y1, xs+x2, ys-y2);	/* lines in eight */
	    GRVector(xs-x1, ys+y1, xs-x2, ys+y2);	/* octets of the */
	    GRVector(xs-x1, ys-y1, xs-x2, ys-y2);	/* circle */
	    GRVector(xs+y1, ys+x1, xs+y2, ys+x2);
	    GRVector(xs+y1, ys-x1, xs+y2, ys-x2);
	    GRVector(xs-y1, ys+x1, xs-y2, ys+x2);
	    GRVector(xs-y1, ys-x1, xs-y2, ys-x2);
	    x1 = x2;
	    y1 = y2;
	}
	return;
    }
	/* else use the more general case for an arc of a circle */
    x1 = cpoint->x;
    y1 = cpoint->y;

    for (i=0; i<extent; ++i) {
	xs -= epsilon * ys;
	x2 = xs + center->x;
	ys += epsilon * xs;
	y2 = ys + center->y;

	GRVector(x1, y1, x2, y2);

	x1 = x2;
	y1 = y2;
    }
}  /* end GRArc */;


/* This routine calculates parameteric values for use in calculating
 * curves.  The values are an approximation of cumulative arc lengths 
 * of the curve (uses cord * length).  For additional information, 
 * see paper cited below.
 */
static
Paramaterize(x, y, h, n)
float x[MAXPOINTS];
float y[MAXPOINTS];
float h[MAXPOINTS];
register n;
{
    register i, j;
    float t1, t2;
    float u[MAXPOINTS];

    n = numpoints;

    for (i=1; i<=n; ++i) {
	u[i] = 0.0;
	for (j=1; j<i; ++j) {
	    t1 = x[j+1] - x[j];
	    t2 = y[j+1] - y[j];
	    u[i] += (float) sqrt((double) ((t1 * t1) + (t2 * t2)));
	}
    }

    for (i=1; i<n; ++i)
	h[i] = u[i+1] - u[i];
}  /* end Paramaterize */


/*
 * This routine solves for the cubic polynomial to fit a spline
 * curve to the the points  specified by the list of values.
 * The curve generated is periodic.  The alogrithms for this 
 * curve are from the "Spline Curve Techniques" paper cited below.
 */
static
PeriodicSpline(h, z, dz, d2z, d3z, npoints)
float h[MAXPOINTS];		/* paramaterization */
float z[MAXPOINTS];		/* point list */
float dz[MAXPOINTS];		/* to return the 1st derivative */
float d2z[MAXPOINTS];		/* 2nd derivative */
float d3z[MAXPOINTS];		/* and 3rd derivative */
register npoints;		/* number of valid points */
{
    float a[MAXPOINTS]; 
    float b[MAXPOINTS]; 
    float c[MAXPOINTS]; 
    float d[MAXPOINTS]; 
    float deltaz[MAXPOINTS];
    float r[MAXPOINTS];
    float s[MAXPOINTS];
    float ftmp;
    register i;

    /* step 1 */
    for (i=1; i<npoints; ++i) {
	if (h[i] != 0)
	    deltaz[i] = (z[i+1] - z[i]) / h[i];
	else
	    deltaz[i] = 0;
    }
    h[0] = h[npoints-1];
    deltaz[0] = deltaz[npoints-1];

    /* step 2 */
    for (i=1; i<npoints-1; ++i) {
	d[i] = deltaz[i+1] - deltaz[i];
    }
    d[0] = deltaz[1] - deltaz[0];

    /* step 3a */
    a[1] = 2 * (h[0] + h[1]);
    if (a[1] == 0) 
	return(-1);  /* 3 consecutive knots at same point */
    b[1] = d[0];
    c[1] = h[0];

    for (i=2; i<npoints-1; ++i) {
	ftmp = h[i-1];
	a[i] = ftmp + ftmp + h[i] + h[i] - (ftmp * ftmp)/a[i-1];
	    if (a[i] == 0) 
		return(-1);  /* 3 consec knots at same point */
	b[i] = d[i-1] - ftmp * b[i-1]/a[i-1];
	c[i] = -ftmp * c[i-1]/a[i-1];
    }

    /* step 3b */
    r[npoints-1] = 1;
    s[npoints-1] = 0;
    for (i=npoints-2; i>0; --i) {
	r[i] = -(h[i] * r[i+1] + c[i])/a[i];
	s[i] = (6 * b[i] - h[i] * s[i+1])/a[i];
    }

    /* step 4 */
    d2z[npoints-1] = (6 * d[npoints-2] - h[0] * s[1] 
		       - h[npoints-1] * s[npoints-2]) 
		     / (h[0] * r[1] + h[npoints-1] * r[npoints-2] 
			+ 2 * (h[npoints-2] + h[0]));
    for (i=1; i<npoints-1; ++i) {
	d2z[i] = r[i] * d2z[npoints-1] + s[i];
    }
    d2z[npoints] = d2z[1];

    /* step 5 */
    for (i=1; i<npoints; ++i) {
	dz[i] = deltaz[i] - h[i] * (2 * d2z[i] + d2z[i+1])/6;
	if (h[i] != 0)
	    d3z[i] = (d2z[i+1] - d2z[i])/h[i];
	else
	    d3z[i] = 0;
    }

    return(0);
}  /* end PeriodicSpline */


/*
 * This routine solves for the cubic polynomial to fit a spline
 * curve from the points specified by the list of values.  The alogrithms for
 * this curve are from the "Spline Curve Techniques" paper cited below.
 */
static
NaturalEndSpline(h, z, dz, d2z, d3z, npoints)
float h[MAXPOINTS];		/* paramaterization */
float z[MAXPOINTS];		/* point list */
float dz[MAXPOINTS];		/* to return the 1st derivative */
float d2z[MAXPOINTS];		/* 2nd derivative */
float d3z[MAXPOINTS];		/* and 3rd derivative */
register npoints;		/* number of valid points */
{
    float a[MAXPOINTS]; 
    float b[MAXPOINTS]; 
    float d[MAXPOINTS]; 
    float deltaz[MAXPOINTS];
    float ftmp;
    register i;

    /* step 1 */
    for (i=1; i<npoints; ++i) {
	if (h[i] != 0)
	    deltaz[i] = (z[i+1] - z[i]) / h[i];
	else
	    deltaz[i] = 0;
    }
    deltaz[0] = deltaz[npoints-1];

    /* step 2 */
    for (i=1; i<npoints-1; ++i) {
	d[i] = deltaz[i+1] - deltaz[i];
    }
    d[0] = deltaz[1] - deltaz[0];

    /* step 3 */
    a[0] = 2 * (h[2] + h[1]);
    if (a[0] == 0)		/* 3 consec knots at same point */
	return(-1);
    b[0] = d[1];

    for (i=1; i<npoints-2; ++i) {
	ftmp = h[i+1];
	a[i] = ftmp + ftmp + h[i+2] + h[i+2] - (ftmp * ftmp) / a[i-1];
	if (a[i] == 0)		/* 3 consec knots at same point */
	    return(-1);
	b[i] = d[i+1] - ftmp * b[i-1]/a[i-1];
    }

    /* step 4 */
    d2z[npoints] = d2z[1] = 0;
    for (i=npoints-1; i>1; --i) {
	d2z[i] = (6 * b[i-2] - h[i] *d2z[i+1])/a[i-2];
    }

    /* step 5 */
    for (i=1; i<npoints; ++i) {
	dz[i] = deltaz[i] - h[i] * (2 * d2z[i] + d2z[i+1])/6;
	if (h[i] != 0)
	    d3z[i] = (d2z[i+1] - d2z[i])/h[i];
	else
	    d3z[i] = 0;
    }

    return(0);
}  /* end NaturalEndSpline */


#define PointsPerInterval 16


/*
 * This routine computes a smooth curve through a set of points.
 * Returns -1 if there are too many knots to draw the curve.
 * Use GRCurve AFTER this routine to actually draw the curve.
 * [Formerly the first half of GRCurve()]
 *
 * The method used is the parametric spline curve on unit knot mesh described
 * in "Spline Curve Techniques" by Patrick Baudelaire, Robert Flegal, and 
 * Robert Sproull -- Xerox Parc.
 */
GRSetCurve(pointlist)
POINT *pointlist;
{
    register POINT *ptr;
    register i, stat;

    /* Copy point list to array for easier access */
    ptr = pointlist;
    for (i=1; (!Nullpoint(ptr)); ++i) {
	x[i] = ptr->x;
	y[i] = ptr->y;
	ptr = PTNextPoint(ptr);
    }

    /* Solve for derivatives of the curve at each point 
       separately for x and y (parametric). */

    numpoints = i - 1;		/* set global numpoints */

    Paramaterize(x, y, h, numpoints);

    stat = 0;
    if ((x[1] == x[numpoints]) && (y[1] == y[numpoints])) { /* closed curve */
	stat |= PeriodicSpline(h, x, dx, d2x, d3x, numpoints);
	stat |= PeriodicSpline(h, y, dy, d2y, d3y, numpoints);
    }
    else {
	stat |= NaturalEndSpline(h, x, dx, d2x, d3x, numpoints);
	stat |= NaturalEndSpline(h, y, dy, d2y, d3y, numpoints);
    }

    curve_set = 1;		/* indicates that paramterization is done */
    return(stat);
}


/*
 * This routine displays a smooth curve through a set of points.  The 
 * method used is the parametric spline curve on unit knot mesh described
 * in "Spline Curve Techniques" by Patrick Baudelaire, Robert Flegal, and 
 * Robert Sproull -- Xerox Parc.
 * [formerly the second half of GRCurve()]
 *
 * Uses the data computed first by GRSetCurve().
 * GRSetCurve() MUST be called before this routine and have returned a ZERO.
 */
/*ARGSUSED*/
GRCurve(style)
int style;
{
    float t, t2, t3;
    float x1, y1, x2, y2;
    register j, k;

    GRSetLineStyle(style);

    x1 = x[1];
    y1 = y[1];

    /* generate the curve using the information from GRSetCurve() and 
       PointsPerInterval vectors between each specified knot. */

    for (j=1; j<numpoints; ++j) {
	for (k=0; k<=PointsPerInterval; ++k) {
	    t = (float) k * h[j] / (float) PointsPerInterval;
	    t2 = t * t;
	    t3 = t2 * t;
	    x2 = x[j] + t * dx[j] + t2 * d2x[j]/2.0 + t3 * d3x[j]/6.0;
	    y2 = y[j] + t * dy[j] + t2 * d2y[j]/2.0 + t3 * d3y[j]/6.0;

	    GRVector(x1, y1, x2, y2);

	    x1 = x2;
	    y1 = y2;
	}
    }
}  /* end GRCurve */


/*
 * This routine clears the Gremlin pix subwindow or current set
 * pixrect image as specified in the mask.
 */
GRClear(mask)
register mask;
{
    XSetFunction(display, pix_gc, GXcopy);
    if (mask & pixmask) {
	XSetForeground(display, pix_gc, background);
	XFillRectangle(display, pix_sw, pix_gc, 0, 0, 
		       (unsigned)pix_size.r_width, (unsigned)pix_size.r_height);
	XSetForeground(display, pix_gc, foreground);
    }

    if (mask & csetmask) {
	XSetForeground(display, pix_gc, background);
	if (cset_rect.r_width != 0) {
	    XFillRectangle(display, cset_pm.pm, pix_gc, 
			   cset_orig.x, cset_orig.y,
			   cset_rect.r_width, cset_rect.r_height);
	}
	XSetForeground(display, pix_gc, foreground);
	cset_rect.r_width = cset_rect.r_height = 0;
    }
}  /* end GRClear */


/*
 *  Display justification of TEXT element.
 */
GRDisplayJustify(elt)
register ELT *elt;
{
    register POINT *point;
    register x, y, length, ysize;

    ysize = GRGetCharYSize(elt->brushf, elt->size);
    length = GRFontStrlen(elt->textpt, elt->brushf, elt->size);
    point = PTNextPoint(elt->ptlist);	/* lower left corner of text */
    x = dbx_to_win(point->x);
    y = dby_to_win(point->y);

    switch (elt->type) {
	case TOPLEFT:
	    y -= ysize;
	    break;
	case TOPCENT:
	    y -= ysize;
	    x += (length >> 1);
	    break;
	case TOPRIGHT:
	    y -= ysize;
	    x += length;
	    break;
	case CENTLEFT:
	    y -= (ysize >> 1);
	    break;
	case CENTCENT:
	    y -= (ysize >> 1);
	    x += (length >> 1);
	    break;
	case CENTRIGHT:
	    y -= (ysize >> 1);
	    x += length;
	    break;
	case BOTLEFT:
	    break;
	case BOTCENT:
	    x += (length >> 1);
	    break;
	case BOTRIGHT:
	    x += length;
	    break;
    }

    XCopyPlane(display, dot_pm.pm, pix_sw, pix_xor_gc, 0, 0, 5, 5, x - 2, y - 2,
		(unsigned long)1);
    XCopyPlane(display, dot_pm.pm, cset_pm.pm, pix_xor_gc, 0, 0, 5, 5, x - 2,
	       y - 2, (unsigned long)1);
}


/*
 * This routine displays a point (layed down by the user) in the
 * pix subwindow.  
 */
GRDisplayPoint(dbx, dby, number)
float dbx, dby;			/* data base coordinates */
register number;		/* point number */
{
    register x, y;
    char numbuf[5];
    XGCValues	gcValues;

    x = dbx_to_win(dbx);
    y = dby_to_win(dby);

    gcValues.font = text_pf->fid;
    XChangeGC(display, pix_xor_gc, GCFont, &gcValues);
    if (Artmode)
	XCopyPlane(display, littlepoint_pm.pm, pix_sw, pix_xor_gc,
		   3, 2, 3, 3, x - 1, y - 1, (unsigned long)1);
    else {
	XCopyPlane(display, littlepoint_pm.pm, pix_sw, pix_xor_gc,
		   1, 7, 7, 7, x - 3, y - 3, (unsigned long)1);
	(void) sprintf(numbuf, "%d", number+1);
	XDrawString(display, pix_sw, pix_xor_gc, x + 5, y + 3,
		    numbuf, strlen(numbuf));
    }
}  /* end GRDisplayPoint */


/*
 * This routine erases the specified point.
 */
GRErasePoint(dbx, dby, number)
float dbx, dby;
register number;
{
    GRDisplayPoint(dbx, dby, number);
}  /* end GRErasePoint */


/*
 * This routine clears all points in plist.
 */
GRBlankPoints(plist)
register POINT *plist;
{
    register i = 0;

    while (!Nullpoint(plist)) {
	GRErasePoint(plist->x, plist->y, i++);
	plist = PTNextPoint(plist);
    }   
}  /* end GRBlankPoints */


/*
 * This routine displays the grid.
 */
GRDisplayGrid()
{
    XGCValues	gcValues;

    gcValues.stipple = grid32_pm.pm;
    gcValues.fill_style = FillStippled;
    XChangeGC(display, pix_xor_gc, GCStipple|GCFillStyle, &gcValues);
    XFillRectangle(display, pix_sw, pix_xor_gc, 0, 0, 
		   pix_size.r_width, pix_size.r_height);
    XSetFillStyle(display, pix_xor_gc, FillSolid);
}  /* end GRDisplayGrid */


/*
 * This routine erases the grid.
 */
GRBlankGrid()
{
    GRDisplayGrid();
}  /* end GRBlankGrid */


/*
 * Flash current set display.
 */
GRCurrentSet()
{
    if (DBNullelt(cset))
	return;

    XCopyArea(display, cset_pm.pm, pix_sw, pix_xor_gc, cset_orig.x, cset_orig.y,
	      cset_rect.r_width, cset_rect.r_height,
	      cset_rect.r_left, cset_rect.r_top);

    CsetOn = !CsetOn;
}


/*
 * Make current set on.
 */
GRCurrentSetOn()
{
    if (!CsetOn)
	GRCurrentSet();
}


/*
 * Make current set off.
 */
GRCurrentSetOff()
{
    if (CsetOn)
	GRCurrentSet();
}


/*
 * Return TRUE if font file exists and is readable.
 */
GRfontfound(font, size)
register font, size;
{
    return(font_info[font-1][size-1] != (XFontStruct *) -1);
}
 

/* 
 * Open the default font file on startup.
 */
GRFontInit()
{
    GROpenFont(CFONT, CSIZE);
    GRStippleInit();
}  /* end GRFontInit */


/*
 * Initialize stipple patterns from font file.
 * Big assumption: all stipples are defined by 32 x 32 bit patterns.
 * All fonts do not contain exactly 32 rows of 4 bytes - this is ok -
 * Fonts wider than 32 bits will be clipped.
 */
GRStippleInit()
{
    register i;

    for (i = 0; i < 32; i++) {
	fill_pms[i].pm = XCreateBitmapFromData(display, root_window,
					       stipple_data[i], 32, 32);
    }

   for (i=0; i< NMENUSTIPPLES; i++) {		    
	XSetFunction(display, pixmap_gc, GXcopy);
	XCopyArea(display, fill_pms[stipple_index[i] - 1].pm,
		  stipple_pms[i]->pm, pixmap_gc,
		  0, 0, 16, 16, 0, 0);
	XSetFunction(display, pixmap_gc, drawOP);
	XCopyArea(display, stipple_pms[i]->pm, menu_stipple_pms[i]->pm,
		  pixmap_gc, 0, 0, 16, 16, 0, 0);
    }
} /* end GRStippleInit */;


/* 
 * Open a font file for use by first reading it into memory.
 * If the file is read successfully, the appropriate entry in
 * font_info[] is set to point to its memory image.  If the
 * file cannot be opened or there is a error in reading the
 * file, set the font_info[] entry to -1.
 */
GROpenFont(font, size)
register font;		/* font is 1 to 4 */
register size;		/* size is 1 to 4 */
{
    char name[256];

    if (font_info[--font][--size] != (XFontStruct *)NULL) {
	/* 
	 * already tried to open.
	 */
	return;
    }

    /*
     * Try a gremlin font.  If that doesn't work, try the
     * normal font.
     */
    sprintf(name, "gremlin-%s-%d",font_types[font], size+1);
    font_info[font][size] = XLoadQueryFont(display, name);

    if (font_info[font][size] == (XFontStruct *)NULL) {
	sprintf(name, font_names[font], font_sizes[size]);
	font_info[font][size] = XLoadQueryFont(display, name);

	if (font_info[font][size] == (XFontStruct *)NULL) {
	    perror(name);
	    font_info[font][size] = (XFontStruct *)-1;
	    return;
	}
    }

    /*
     * Want the height of this font.  Unfortunately the converted fonts
     * think they are bigger than they really are so I have to scale the 
     * height down.  If we get better fonts then we won't have this
     * problem.
     */
    charysizes[font][size] = (font_info[font][size]->max_bounds.ascent +
			      font_info[font][size]->max_bounds.descent) * 0.7;
}  /* end GROpenFont */


/*
 * Determine pixel length of string s in font and size.
 */
GRFontStrlen(text, font, size)
char *text;
int font;		/* 1 - 4 */
int size;		/* 1 - 4 */
{
    if (*text == '\0')
	return(0);

    if (font_info[font-1][size-1] == (XFontStruct *)NULL) /* not open yet */
	GROpenFont(font, size);

    if (!GRfontfound(font, size))		/* unreadable font */
	return(0);

    return(XTextWidth(font_info[font-1][size-1], text, strlen(text)));
}

/*
 * Display text string of font/size at position pos.
 */
GRPutText(text, font, size, pos)
char *text;
int font, size; 
POINT *pos;
{
    XGCValues	gcValues;
    int		x, y;
    int		len;

    if (font_info[font-1][size-1] == NULL)
	GROpenFont(font, size);

    if (!GRfontfound(font, size))
	return;

    gcValues.font = font_info[font-1][size-1]->fid;
    XChangeGC(display, pix_xor_gc, GCFont, &gcValues);

    len = strlen(text);

    x = dbx_to_win(pos->x);
    y = dby_to_win(pos->y);
    XDrawString(display, scratch_pm.pm, pix_xor_gc, x, y, text, len);
    minsunx = x;
    maxsuny = y + 8;
    minsuny = y = GRGetCharYSize(font, size);
    maxsunx = x + XTextWidth(font_info[font-1][size-1], text, len);


} /* end GRPutText */;


/* 
 * Set the actual positioning point for text with the justify, font, and
 * size attributes.  Point is a pointer to the POINT layed down by the user.
 * Pos is a pointer to the returned positioning POINT used in a subsequent
 * call to GRPutText(). 
 */
GRSetTextPos(text, justify, font, size, point, pos)
char *text;
int justify, font, size; 
POINT *point, *pos;
{
    register length;
    register charysize;

    charysize = GRGetCharYSize(font, size);
    length = GRFontStrlen(text, font, size);

    switch (justify) {
	case BOTLEFT:	
	    pos->x = point->x;
	    pos->y = point->y;
	    break;
	case BOTCENT:	
	    pos->x = point->x - (length / 2);
	    pos->y = point->y;
	    break;
	case BOTRIGHT:	
	    pos->x = point->x - length;
	    pos->y = point->y;
	    break;
	case CENTLEFT:	
	    pos->x = point->x;
	    pos->y = point->y - (charysize / 2);
	    break;
	case CENTCENT:	
	    pos->x = point->x - (length / 2);
	    pos->y = point->y - (charysize / 2);
	    break;
	case CENTRIGHT:	
	    pos->x = point->x - length;
	    pos->y = point->y - (charysize / 2);
	    break;
	case TOPLEFT:	
	    pos->x = point->x;
	    pos->y = point->y - charysize;
	    break;
	case TOPCENT:	
	    pos->x = point->x - (length / 2);
	    pos->y = point->y - charysize;
	    break;
	case TOPRIGHT:	
	    pos->x = point->x - length;
	    pos->y = point->y - charysize;
	    break;
    }
}
