/*
 * @(#)help.c	1.2	1/3/85
 *
 * Routines to provide help screens for the SUN Gremlin picture editor.
 *
 * Mark Opperman (opcode@monet.BERKELEY)
 *
 * Modifications by Jim Thompson (james@ernie.berkeley.edu)
 */

#include "gremlin.h"
#include "icondata.h"

/* imports from main.c */

extern Window		pix_sw;
extern struct rect 	pix_size;
extern PMRec		scratch_pm;
extern XFontStruct	*text_pf;
extern Window		menu_sw;
extern Window		text_sw;

/* imports from sun.c */

extern get_any_button();
static help_message();

/* locals */

static help_x, help_y;		/* current position on help screen */

extern unsigned text_height;

#define help_dy  (text_height - 2)
#define HELP_LEFT 8
#define HELP_SKIP 16

/*
 * Display a help screen consisting of a pixrect, a title and some text.
 * The current picture is saved in the scratch pixrect before displaying
 * the help screen.  After the user presses a mouse button, the picture
 * is restored and processing continues.
 */
help_screen(icon_pm, title, text)
PMRec *icon_pm;
char *title, *text;
{
    int clear_scratch();

    /* save current display */
    XSetFunction(display, pix_gc, GXcopy);
    XCopyArea(display, pix_sw, scratch_pm.pm, pix_gc,
	      0, 0, pix_size.r_width, pix_size.r_height, 0, 0);
    /* clear picture subwindow before help screen display */
    XSetForeground(display, pix_gc, background);
    XFillRectangle(display, pix_sw, pix_gc, 0, 0, pix_size.r_width,
		   pix_size.r_height);
    XSetForeground(display, pix_gc, foreground);

    help_x = HELP_LEFT;
    help_y = 8;

    /* display icon if not a NULL pointer */
    if (icon_pm != NULL) {
	XCopyPlane(display, icon_pm->pm, pix_sw, pix_gc,
		  0, 0, icon_pm->width, icon_pm->height, help_x, help_y, 
		  (unsigned long)1);
	help_x += icon_pm->width + HELP_SKIP;
	help_message(title);
	help_y += icon_pm->height + HELP_SKIP;
    } else {
	help_message(title);
	help_y += help_dy + HELP_SKIP;
    }

    help_x = HELP_LEFT;
    help_message(text);

    help_y += HELP_SKIP;
    help_x = HELP_LEFT;
    help_message("Press a mouse button to continue.");

    /* wait for mouse button in any subwindow or tool border */
    get_any_button();

    /* restore picture subwindow */
    XSetFunction(display, pix_gc, GXcopy);
    XCopyArea(display, scratch_pm.pm, pix_sw, pix_gc,
	      0, 0, pix_size.r_width, pix_size.r_height, 0, 0);
    clear_scratch();
}


/*
 * A simple text display tool for help messages.
 * Text is displayed at the current help location (help_x, help_y).
 * Newlines within the text string are handled appropriately.
 * Maximum line length (between newlines) is 80 characters, although 
 * only about 64 can be displayed on the default Gremlin screen.
 */
static
help_message(text)
char *text;
{
    char buf[80];
    register i;
    XGCValues	gcValues;

    while (*text != '\0') {
	i = 0;

	while ((*text != '\n') && (*text != '\0'))
	    buf[i++] = *text++;
	buf[i] = '\0';

	if (*text == '\n')
	    text++;

	gcValues.font = text_pf->fid;
	XChangeGC(display, pix_xor_gc, GCFont, &gcValues);
	XDrawString(display, pix_sw, pix_xor_gc, 
	            help_x, help_y + TEXT_BASELINE + 2, buf, strlen(buf));
	help_y += help_dy;
    }
}


static char help_HELP[] = "\
The three subwindows in Gremlin are used for text entry,\n\
menu display and command selection, and picture display.\n\
In addition, the normal tool manager pop-up menu is available\n\
from the tool borders.\n\
\n\
In the menu subwindow, the left mouse button chooses commands by \n\
selecting the appropriate icon, the middle mouse button does the\n\
same command on only the current set (where appropriate), and\n\
the right button provides a help screen for the icon.\n\
\n\
Many icons have a 'short' single-key equivalent. These are listed\n\
after the command title on the corresponding help screen.\n\
For more information about short commands, select the Help icon with\n\
the middle button.\n\
\n\
Most drawing commands can also be performed interactively\n\
by selecting the menu icon before placing points in the\n\
picture subwindow.  The icon is highlighted, and remains in\n\
effect until the same or another interactive icon is selected.\n\
\n\
For more information on the picture subwindow, press the right mouse\n\
button in the picture and select the Help option\n\
\n\
For more information about the text subwindow, press the right mouse\n\
button in the text window";

help()
{
    help_screen(&question_pm, "Help ('?')", help_HELP);
}


static char short_HELP[] = "\
Short commands are single-key commands which correspond to menu icons.\n\
They can be typed with the cursor in the picture or menu subwindow, and\n\
respond just as if the corresponding icon were selected with the left\n\
mouse button. The short commands are:\n\
\n\
   1-4 - Save in Buffer 1-4            q - Toggle Grid\n\
     a - Draw arc/circle               r - Rotate current set\n\
     b - Draw curve (b-spline)         s - Scale (XY) current set\n\
     c - Copy current set              t - Translate (move) current set\n\
     d - Define set by points          u - Undo\n\
     e - Erase current set             v - Draw Vector\n\
     f - Define Current Set by area    w - Draw Arrow\n\
     g - Toggle Gravity                x - Draw Box\n\
     h - Hide window                   z - Toggle Manhattan Adjustment\n\
     i - Make Iconic                   ? - Help\n\
     o - Window On Top            ctrl-L - Redisplay screen\n\
     n - No interactive mode           . - Repeat previous command\n\
\n\
The SHIFT key is a special short command which suspends the current mode\n\
and enters the interactive Define Current Set mode.  When the SHIFT key is\n\
released, the previous function resumes.";

short_help()
{
    help_screen(&question_pm, "Short Commands", short_HELP);
}


static char textsw_HELP[] = "\
Text and command arguments (when required) are entered in this\n\
subwindow from the keyboard.  The quick form of TEXT entry is invoked\n\
by pressing RETURN after entering a string to be displayed in the\n\
picture.  The string is displayed at the FIRST point placed (using\n\
the current justification mode), and this point is removed from the\n\
display.  All other points are renumbered.\n\
\n\
The middle mouse button is used to redisplay the previous text string.\n\
The left mouse button will display the text string from the current set.\n\
\n\
Simple editing commands can be used to modify the text. These are\n\
backspace and line erase (usually ctrl-h and ctrl-u) and\n\
word erase (ctrl-w)\n\
\n\
All text insertion and editing occurs at the cursor location. The cursor\n\
can be moved by selecting the new location with the left mouse button.\n\
It can also be moved forward and backward using ctrl-f and ctrl-b.\n\
\n\
Arguments to commands such as Read must be entered before the command is\n\
selected.";

textsw_help()
{
    help_screen((PMRec *)NULL, "Text Subwindow Help", textsw_HELP);
}


static char pixsw_HELP[] = "\
The left mouse button is used to lay down points.  The middle\n\
mouse button erases points in the opposite order from which\n\
they were layed down.\n\
\n\
When interactive, the left button generally is used to specify\n\
points, with a double-click on the last point.  The center button\n\
removes points, or acts as an UNDO button.  See individual icon\n\
help for specific instructions.\n\
\n\
The following general rules apply while interactive:\n\
    a. An approximation of the figure will be displayed according\n\
       to the current position of the mouse.\n\
    b. All alignments, adjustments, and gravity are in effect.\n\
       However, gravity is only visible when the mouse pauses.\n\
    c. If the cursor leaves the picture subwindow while performing\n\
       an interactive action, the picture reverts to its state\n\
       before the action.  This provides a convenient \"escape\" if\n\
       you don't like what you see.\n\
    d. The right button displays a menu of common commands. ";

pixsw_help()
{
    help_screen((PMRec *)NULL, "Picture Subwindow Help", pixsw_HELP);
}


menusw_help()
{
    /* keeps getting invoked by mistake - not useful anyway */
}

clear_scratch()
{
    XSetFunction(display, pix_gc, GXcopy);
    XSetForeground(display, pix_gc, background);
    XFillRectangle(display, scratch_pm.pm, pix_gc, 0, 0, root_width,
		   root_height);
    XSetForeground(display, pix_gc, foreground);
} 

