/*
 * spline.c
 *
 * Spline routines for the SUN Gremlin picture editor,
 * Two types of splines are supported
 * 	(1) B-splines (B_SPLINE) : uniform cubic B-spline,
 *	(2) Bezier Cubics (BEZIER).
 *
 * Author: Peehong Chen (phc@renoir.berkeley.edu)
 * Date: 9/17/1986
 *
 */

#include <stdio.h>
#include "gremlin.h"

/* Imports from graphics.c */

extern	GRVector();
extern	GRSetLineStyle();
extern	curve_set;

/* Local Variables */

static int numpoints;
static float Qx, Qy;
static float x[MAXPOINTS], y[MAXPOINTS];

/*
 * B_SPLINE Routines
 */

#define PointsPerInterval 	16
#define Delta_U			1.0/PointsPerInterval


/*
 * This routine computes a point in B-spline segment, given i, and u.
 * Details of this algorithm can be found in the tech. report cited below.
 */
static
UniformCubicBSpline(i, u)
int i;
float u;
{
    float u2, u3, b_2, b_1, b0, b1;
    register i1, i_2, i_1;

    i1  = i + 1;
    i_1 = i - 1;
    i_2 = i - 2;

    u2 = u * u;
    u3 = u2 * u;
    b_2 = (1 - 3*u + 3*u2 - u3) / 6.0;
    b_1 = (4 - 6*u2 + 3*u3) / 6.0;
    b0  = (1 + 3*u + 3*u2 - 3*u3) / 6.0;
    b1  = u3 / 6.0;

    Qx = b_2 * x[i_2] + b_1 * x[i_1] + b0 * x[i] + b1 * x[i1];
    Qy = b_2 * y[i_2] + b_1 * y[i_1] + b0 * y[i] + b1 * y[i1];	
} /* end UniformCubicBSpline */


/*
 * This routine initializes the array of control vertices
 * We consider two end conditions here:
 *   (1) closed curve -- C2 continuation and end vertex not interpolated, i.e. 
 *		V[0] = V[n-1], and
 *		V[n+1] = V[2].
 *   (2) open curve -- end vertex interpolation, i.e.
 *		V[0] = 2*V[1] - V[2], and
 *		V[n+1] = 2*V[n] - V[n-1].
 * Details of these two and other end conditions can be found in tech. report
 * cited below.
 */
GRSetBSpline(pointlist)
POINT *pointlist;
{
    register POINT *ptr;
    register n_1, n, n1;

    /* Copy point list to array for easier access */
    ptr = pointlist;
    for (n=1; (!Nullpoint(ptr)); ++n) {
	x[n] = ptr->x;
	y[n] = ptr->y;
	ptr = PTNextPoint(ptr);
    }

    if (n > MAXPOINTS) {		 /* Overflow*/
	return(-1);
    } else {
	numpoints = --n;
    }

    n_1 = n - 1;
    n1  = n + 1;
  
    if ((x[1] == x[numpoints]) && (y[1] == y[numpoints])) { /* closed curve */
	x[0] = x[n_1];				/* V[0] */
	y[0] = y[n_1];
	x[n1] = x[2];				/* V[n+1] */
	y[n1] = y[2];
    }
    else {				/* end vertex interpolation */
	x[0] = 2*x[1] - x[2];			/* V[0] */
	y[0] = 2*y[1] - y[2];
	x[n1] = 2*x[n] - x[n_1];		/* V[n+1] */
	y[n1] = 2*y[n] - y[n_1];
    }

    curve_set = 1;		/* indicates that paramterization is done */
    return(0);
} /* end GRSetBSpline */


/*
 * This routine renders a uniform cubic B-Spline.
 * Uniform means knot spacing is assumed to be of unit length.
 * Two important features of B-Splines are
 * 	(1) convex hull property: the curve is smoother than a
 *	    cubic interpolatory spline, given the same set of points.
 *	(2) local control: changing the shape of one local region of
 *	    the curve does not affect the shape of remote portions.
 * Details of uniform cubic B-Splines can be found in Chapters 4-5 of
 * Richard H. Bartels and Brian A. Barsky,
 * "An Introduction to the Use of Splines in Computer Graphics",
 * Tech. Report CS-83-136, Computer Science Division,
 * University of California, Berkeley, 1984.
 */
GRBSpline(style)
int style;
{
    float x0, y0, u;
    register i;

    GRSetLineStyle(style);

    if (numpoints < 2) {
	return(0);
    }
    if (numpoints == 2) {
	GRVector(x[1], y[1], x[2], y[2]);
	return(0);
    }

    UniformCubicBSpline(2, 0.0);
    x0 = Qx;
    y0 = Qy;
    
    for (i = 2; i <= numpoints; i++) {
	for (u = Delta_U; u <= 1.0; u += Delta_U) {
	    UniformCubicBSpline(i, u);
	    GRVector(x0, y0, Qx, Qy);
	    x0 = Qx;
	    y0 = Qy;
	}
    }
    return(0);
} /* end GRBSpline */


/*
 * This routine renders a segment of Bezier cubics.
 *	Q(u) = V[1]*(1-u)^3 + V[2]*3*u*(1-u)^2 + V[3]*3*u^2*(1-u) + V[4]*u^3
 *	     = (-V[1]+3*(V[2]-V[3])+V[4])*u^3 + 3*(V[1]-2*V[2]+V[3])*u^2
 *	       - 3*(V[1]-V[2])*u + V[1]
 */
static
BezierCubics(x1, y1, x2, y2, x3, y3, x4, y4)
float x1, y1, x2, y2, x3, y3, x4, y4;
{
    float bx, by, cx, cy, dx, dy, xi, yi, xj, yj, u, u2, u3;
    
    bx = -3 * (x1-x2);
    by = -3 * (y1-y2);
    cx = 3 * (x1 - 2*x2 + x3);
    cy = 3 * (y1 - 2*y2 + y3);
    dx = -x1 + 3*(x2-x3) + x4;
    dy = -y1 + 3*(y2-y3) + y4;

    xi = x1;
    yi = y1;
    for (u = Delta_U; u <= 1.0; u += Delta_U) {
	u2 = u * u;
	u3 = u2 * u;
	xj = x1 + bx * u + cx * u2 + dx * u3;
	yj = y1 + by * u + cy * u2 + dy * u3;
	GRVector(xi, yi, xj, yj);
	xi = xj;
	yi = yj;
    }
} /*end BezierCubics */


/*
 * This routine initializes an array to a list of offsets, each of which
 * is the offset from V[i] to V[i+1], that is
 * 	O[i] = (x[i], y[i]) = V[i+1] - V[i],
 * for i=1 to N-1, where N is the number of points given.
 * The starting end point (V[1]) is saved in (Qx, Qy).
 */
GRSetBezier(pointlist)
POINT *pointlist;
{
    register	POINT	*ptr;
    register	n;
    float	Lx, Ly;			/* last point */

    /* Copy offset list to array for easier access */
    ptr = pointlist;
    Qx = Lx = ptr->x;
    Qy = Ly = ptr->y;
    ptr = PTNextPoint(ptr);
    for (n=1; (!Nullpoint(ptr)); ++n) {
	x[n] = ptr->x - Lx;
	y[n] = ptr->y - Ly;
	Lx = ptr->x;
	Ly = ptr->y;
	ptr = PTNextPoint(ptr);
    }

    if (n > MAXPOINTS) {		 /* Overflow*/
	return(-1);
    } else {
	numpoints = n;
    }

    curve_set = 1;		/* indicates that paramterization is done */
    return(0);
} /* end GRSetBezier */


/*
 * This routine draws piecewise continuous Bezier cubics.
 * Given 2 points, a stringht line is drawn.
 * Given 3 points V[1], V[2], and V[3], a Bezier cubic segment
 * of (V[1], (V[1]+V[2])/2, (V[2]+V[3])/2, V[3]) is drawn.
 * In the case when N (N >= 4) points are given, N-2 Bezier segments will
 * be drawn, each of which (for i=1 to N-2) is 
 *	(Q, Q+O[i]/3, Q+(3*O[i]+O[i+1])/6, K+O[i+1]/2),
 * where
 *	Q is the current point,
 *	K is the continuation offset = Qinitial + Sigma(1, i)(O[i])
 * Note that when i is 1, the initial point
 *	Q = V[1].
 * and when i is N-2, the terminating point
 *	K+O[i+1]/2 = V[N].
 */
GRBezier(style)
int style;
{
    register	i, i1;
    float	x1, y1, x2, y2, x3, y3, Kx, Ky;

    GRSetLineStyle(style);

    if (numpoints < 2) {
	return(0);
    }
    if (numpoints == 2) {
	GRVector(Qx, Qy, Qx+x[1], Qy+y[1]);
	return(0);
    }
    if (numpoints == 3) {
	x1 = Qx + x[1];
	y1 = Qy + y[1];
	x2 = x1 + x[2];
	y2 = y1 + y[2];
        BezierCubics(Qx, Qy, (Qx+x1)/2.0, (Qy+y1)/2.0,
		     (x1+x2)/2.0, (y1+y2)/2.0, x2, y2);
	return(0);
    }
  
    /* numpoints >= 4 */
    Kx = Qx + x[1];
    Ky = Qy + y[1];
    x[1] = 2 * x[1];
    y[1] = 2 * y[1];
    i = numpoints - 1;
    x[i] = 2 * x[i];
    y[i] = 2 * y[i];
    i1 = 2;
    for (i = 1; i <= numpoints-2; i++) {
	x1 = Qx + x[i]/3;
	y1 = Qy + y[i]/3;
	x2 = Qx + (3*x[i] + x[i1])/6;
	y2 = Qy + (3*y[i] + y[i1])/6;
	x3 = Kx + x[i1]/2;
	y3 = Ky + y[i1]/2;
	BezierCubics(Qx, Qy, x1, y1, x2, y2, x3, y3);
	Qx = x3;
	Qy = y3;
	Kx = Kx + x[i1];
	Ky = Ky + y[i1];
	i1++;
    }
    return(0);
} /* end GRBezier */
